import pathlib
import sys
import typing

import pytest
import tree_sitter as ts

import tree_sitter_type_provider as tstp


def node_dict_simplify(node_dict: dict[str, typing.Any]) -> None:
    if len(node_dict) > 4:
        del node_dict["text"]

    del node_dict["start_position"]
    del node_dict["end_position"]

    for key in node_dict.keys():
        if isinstance(node_dict[key], dict):
            node_dict_simplify(node_dict[key])
        if isinstance(node_dict[key], list):
            for val in node_dict[key]:
                if isinstance(val, dict):
                    node_dict_simplify(val)


TESTDIR = pathlib.Path(__file__).parent


@pytest.mark.golden_test("golden/convert/*.yml")
def test_talon(golden):
    class_prefix = golden["input"]["class_prefix"]
    module_name = f"tree_sitter_{ golden['input']['name'] }"

    def as_class_name(node_type_name: str) -> str:
        buffer: list[str] = [class_prefix]
        for part in node_type_name.split("_"):
            buffer.append(part.capitalize())
        return "".join(buffer)

    node_types_json = TESTDIR / golden["input"]["node_types"]
    node_types = tstp.NodeType.schema().loads(node_types_json.read_text(), many=True)

    repository_path = str(TESTDIR / golden["input"]["repository_path"])
    library_name = {
        "linux": f"{ module_name }.so",
        "darwin": f"{ module_name }.dylib",
        "win32": f"{ module_name }.dll",
    }[sys.platform]
    library_path = str(TESTDIR / library_name)
    ts.Language.build_library(library_path, [repository_path])

    language = ts.Language(library_path, golden["input"]["name"])
    parser = ts.Parser()
    parser.set_language(language)

    types = tstp.TreeSitterTypeProvider(
        module_name,
        node_types,
        error_as_node=True,
        as_class_name=as_class_name,
        extra=golden["input"]["extra"],
    )

    contents = golden["input"]["contents"]
    tree = parser.parse(bytes(contents, encoding="utf-8"))
    node = types.from_tree_sitter(tree.root_node)
    node_dict = node.to_dict()
    node_dict_simplify(node_dict)

    assert node_dict == golden.out["output"]



import collections.abc
import dataclasses
import functools
import typing

import dataclasses_json


class NodeTypeError(Exception):
    pass


@dataclasses.dataclass
class Point:
    row: int
    column: int


NodeTypeName = str

NodeFieldName = str

AsClassName = collections.abc.Callable[[NodeTypeName], str]


@dataclasses.dataclass
class Node(dataclasses_json.DataClassJsonMixin):
    text: str
    type_name: NodeTypeName = dataclasses.field(
        metadata=dataclasses_json.config(field_name="type")
    )
    start_position: Point
    end_position: Point


@dataclasses.dataclass
class Leaf(Node):
    pass


@dataclasses.dataclass
class Branch(Node):
    children: typing.Union[None, Node, typing.Sequence[Node]]


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class SimpleNodeType:
    type_name: NodeTypeName = dataclasses.field(
        metadata=dataclasses_json.config(field_name="type")
    )
    named: bool

    def is_extra(self, *, extra: typing.Sequence["SimpleNodeType"]) -> bool:
        return any(self.type_name == other.type_name for other in extra)

    def as_typehint(self, *, as_class_name: AsClassName) -> type[Node]:
        if self.named:
            return typing.cast(type, as_class_name(self.type_name))
        raise ValueError(self)

    @staticmethod
    def many_as_typehint(
        simple_node_types: typing.Sequence["SimpleNodeType"],
        *,
        as_class_name: AsClassName,
    ) -> typing.Optional[type[Node]]:
        Ts: list[type] = []
        for simple_node_type in simple_node_types:
            if simple_node_type.named:
                Ts.append(simple_node_type.as_typehint(as_class_name=as_class_name))

        if len(Ts) == 0:
            return None

        if len(Ts) == 1:
            return Ts[0]

        return functools.reduce(lambda R, T: typing.cast(type, typing.Union[R, T]), Ts)


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class NodeArgsType:
    multiple: bool = False
    required: bool = False
    types: list[SimpleNodeType] = dataclasses.field(default_factory=list)

    @property
    def named(self) -> bool:
        return any(type.named for type in self.types)

    def __bool__(self) -> bool:
        return bool(self.types)

    def as_typehint(
        self,
        *,
        is_field: bool,
        as_class_name: AsClassName,
        extra: typing.Sequence[SimpleNodeType],
    ) -> typing.Optional[type[Node]]:
        if is_field:
            types_with_extra = tuple(self.types)
            multiple_with_extra = self.multiple
        else:
            # Add the extra nodes into the possible children
            types_with_extra = tuple((*self.types, *extra))
            # If there are extra nodes, a node can have any number of children
            multiple_with_extra = bool(extra) or self.multiple
        T = SimpleNodeType.many_as_typehint(
            types_with_extra, as_class_name=as_class_name
        )
        if T is not None:
            if multiple_with_extra:
                return list[T]  # type: ignore
            else:
                if self.required:
                    return T
                else:
                    return typing.Optional[T]  # type: ignore
        else:
            return None


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class NodeType(SimpleNodeType):
    fields: dict[NodeFieldName, NodeArgsType] = dataclasses.field(default_factory=dict)
    children: NodeArgsType = dataclasses.field(default_factory=NodeArgsType)
    subtypes: list[SimpleNodeType] = dataclasses.field(default_factory=list)

    def __post_init__(self, **kwargs):
        assert not (
            self.is_abstract and self.has_content
        ), "Nodes can have either fields and children or subtypes, but not both."

    @property
    def is_abstract(self) -> bool:
        return len(self.subtypes) > 0

    @property
    def has_content(self) -> bool:
        return len(self.fields) > 0 or bool(self.children)

    def as_type(
        self,
        *,
        as_class_name: AsClassName,
        mixins: typing.Sequence[type] = (),
        extra: typing.Sequence[SimpleNodeType],
        **kwargs,
    ) -> type[Node]:
        if self.named:
            cls_name = as_class_name(self.type_name)
            if self.is_abstract:
                # TODO: should be a dynamic type alias
                return type(cls_name, (Node,), {})
            else:
                fields: dict[NodeFieldName, type] = {}

                # Create fields for dataclass
                for field_name, field in self.fields.items():
                    if field.named:
                        field_type = field.as_typehint(
                            is_field=True, as_class_name=as_class_name, extra=extra
                        )
                        if field_type is not None:
                            fields[field_name] = field_type

                # Create children for dataclass
                if self.has_content:
                    children_type = self.children.as_typehint(
                        is_field=False, as_class_name=as_class_name, extra=extra
                    )
                    if children_type:
                        fields["children"] = children_type
                    else:
                        fields["children"] = typing.cast(type, None)

                # Create bases for dataclass
                base: type[Node]
                if self.is_abstract:
                    base = Node
                elif self.has_content:
                    base = Branch
                else:
                    base = Leaf

                # Create and return dataclass
                return dataclasses.make_dataclass(
                    cls_name=cls_name,
                    fields=fields.items(),
                    bases=(base, *mixins),
                    **kwargs,
                )
        else:
            raise ValueError(self)



from __future__ import annotations

import collections.abc
import inspect
import pathlib
import re

import pytest

import tree_sitter_type_provider


def short(sig: inspect.Signature) -> str:
    out = str(sig)
    out = out.replace("tree_sitter_type_provider.node_types.", "")
    out = re.sub(r"ForwardRef\('(\w+)'\)", r"\1", out)
    out = out.replace("typing.", "")
    out = out.replace("NoneType", "None")
    out = out.replace("abc.", "")
    out = out.replace("~Result", "Result")
    return out


def function_signatures(object: object) -> collections.abc.Iterator[str]:
    for name, fun in inspect.getmembers(object, inspect.isfunction):
        if not (name.startswith("_") or name in ["to_dict", "to_json"]):
            try:
                funsig = inspect.signature(fun)
                yield f"{name}{short(funsig)}"
            except ValueError:
                pass


def class_signatures(object: object) -> collections.abc.Iterator[str]:
    for name, cls in inspect.getmembers(object, inspect.isclass):
        if not name.startswith("_"):
            try:
                clssig = inspect.signature(cls)
                yield f"{name}{short(clssig)}"
                for funsigstr in function_signatures(cls):
                    yield f"  {funsigstr}"
            except ValueError:
                pass


@pytest.mark.golden_test("golden/api/*.yml")
def test_talon(golden):
    class_prefix = golden["input"]["class_prefix"]
    module_name = f"tree_sitter_{ golden['input']['name'] }"

    def as_class_name(node_type_name: str) -> str:
        buffer: list[str] = [class_prefix]
        for part in node_type_name.split("_"):
            buffer.append(part.capitalize())
        return "".join(buffer)

    node_types_json = pathlib.Path(__file__).parent / golden["input"]["file"]
    node_types = tree_sitter_type_provider.NodeType.schema().loads(
        node_types_json.read_text(), many=True
    )

    module = tree_sitter_type_provider.TreeSitterTypeProvider(
        module_name,
        node_types,
        error_as_node=True,
        as_class_name=as_class_name,
        extra=golden["input"]["extra"],
    )

    globals().update(module.__dict__)

    output: list[str] = []
    output.extend(function_signatures(module.__class__))
    output.extend(function_signatures(module))
    output.extend(class_signatures(module.__class__))
    output.extend(class_signatures(module))

    assert "\n".join(output) == golden.out["output"]



import collections.abc
import types
import typing

import tree_sitter

from tree_sitter_type_provider.node_types import Branch as Branch
from tree_sitter_type_provider.node_types import Leaf as Leaf
from tree_sitter_type_provider.node_types import Node as Node
from tree_sitter_type_provider.node_types import NodeArgsType as NodeArgsType
from tree_sitter_type_provider.node_types import NodeFieldName as NodeFieldName
from tree_sitter_type_provider.node_types import NodeType as NodeType
from tree_sitter_type_provider.node_types import NodeTypeError as NodeTypeError
from tree_sitter_type_provider.node_types import NodeTypeName as NodeTypeName
from tree_sitter_type_provider.node_types import Point as Point


class TreeSitterTypeProvider(types.ModuleType):
    @staticmethod
    def _tspoint_to_point(tspoint: tuple[int, int]) -> Point:
        return Point(row=tspoint[0], column=tspoint[1])

    def _node_type_name(
        self,
        node: typing.Union[NodeTypeName, Node, NodeType, tree_sitter.Node],
    ) -> str:
        if isinstance(node, tree_sitter.Node):
            return node.type
        elif isinstance(node, Node):
            return node.type_name
        elif isinstance(node, NodeType):
            return node.type_name
        elif isinstance(node, str):
            return node
        else:
            raise NodeTypeError(
                f"Cannot get node type name for object of type {type(node)}", node
            )

    def _node_type(
        self,
        node: typing.Union[NodeTypeName, Node, NodeType, tree_sitter.Node],
    ) -> NodeType:
        node_type_name = self._node_type_name(node)
        node_type = self._node_types_by_type.get(node_type_name, None)
        if node_type is None:
            raise NodeTypeError(f"Could not find node type {node_type_name}")
        return node_type

    def _node_has_children(
        self,
        node: typing.Union[NodeTypeName, Node, NodeType, tree_sitter.Node],
    ) -> bool:
        node_type = self._node_type(node)
        # NOTE: Any node with content can have extra nodes,
        #       even if the node only has fields and no children.
        return node_type.has_content

    def _node_class(
        self, node: typing.Union[Node, NodeType, tree_sitter.Node]
    ) -> type[Node]:
        node_type_name = self._node_type_name(node)
        cls = self._node_classes_by_type.get(node_type_name, None)
        if cls is None:
            raise NodeTypeError(f"Could not find node type {node_type_name}")
        return cls

    def _node_hash(self, tsnode: tree_sitter.Node) -> int:
        return hash((tsnode.start_byte, tsnode.end_byte))

    def from_tree_sitter(
        self,
        tsvalue: typing.Union[
            tree_sitter.Tree, tree_sitter.Node, tree_sitter.TreeCursor
        ],
        *,
        encoding: str = "utf-8",
    ) -> Node:
        if isinstance(tsvalue, tree_sitter.Tree):
            tsvalue = tsvalue.root_node
        if isinstance(tsvalue, tree_sitter.Node):
            tsvalue = tsvalue.walk()
        return self._from_tree_cursor(tsvalue, encoding=encoding)

    def _from_tree_cursor(
        self,
        tscursor: tree_sitter.TreeCursor,
        *,
        encoding: str = "utf-8",
    ) -> Node:
        if tscursor.node.is_named:
            # Convert basic information
            text: str = tscursor.node.text.decode(encoding)
            type_name: str = tscursor.node.type
            start_position: Point = self._tspoint_to_point(tscursor.node.start_point)
            end_position: Point = self._tspoint_to_point(tscursor.node.end_point)

            # Convert children
            fields: dict[str, Node] = {}
            children: list[Node] = []

            def convert_child(tscursor: tree_sitter.TreeCursor):
                field_name = tscursor.current_field_name()
                child = self._from_tree_cursor(tscursor, encoding=encoding)
                if field_name is None:
                    children.append(child)
                else:
                    fields[field_name] = child

            if tscursor.goto_first_child():
                if tscursor.node.is_named:
                    convert_child(tscursor)
                while tscursor.goto_next_sibling():
                    if tscursor.node.is_named:
                        convert_child(tscursor)
                assert tscursor.goto_parent()

            # Create node instance
            kwargs: dict[str, typing.Union[str, Point, Node, list[Node]]] = {}
            kwargs["type_name"] = type_name
            kwargs["text"] = text
            kwargs["start_position"] = start_position
            kwargs["end_position"] = end_position
            if self._node_has_children(tscursor.node):
                kwargs["children"] = children
            kwargs |= fields

            # Return the node with its field name
            return self._node_class(tscursor.node)(**kwargs)  # type: ignore
        else:
            raise TypeError(tscursor.node.type)

    def __init__(
        self,
        module_name: str,
        node_types: typing.Sequence[NodeType],
        *,
        error_as_node: bool = False,
        extra: typing.Sequence[NodeTypeName] = (),
        as_class_name: typing.Optional[collections.abc.Callable[[str], str]] = None,
        mixins: typing.Sequence[type] = (),
        dataclass_kwargs: dict[str, typing.Any] = {},
    ):
        super().__init__(name=module_name)

        if as_class_name is None:

            def snake_to_pascal(node_type_name: str) -> str:
                return "".join(map(str.capitalize, node_type_name.split("_")))

            as_class_name = snake_to_pascal

        # Dictionary of named node types
        self._node_types_by_type: dict[str, NodeType] = {}
        for node_type in node_types:
            if node_type.named:
                self._node_types_by_type[node_type.type_name] = node_type

        # Create a node type for ERROR
        if error_as_node:
            self._node_types_by_type["ERROR"] = NodeType(
                type_name="ERROR",
                named=True,
                children=NodeArgsType(
                    multiple=True,
                    required=True,
                    types=list(node_types),
                ),
            )

        # Create the list of extra node types
        extra_node_types = list(self._node_type(type_name) for type_name in extra)
        if error_as_node and "ERROR" not in extra:
            extra_node_types.append(self._node_type("ERROR"))
        self._extra: typing.Sequence = tuple(extra_node_types)

        # Dictionary of abstract classes
        self._node_bases_by_type: dict[str, list[type[Node]]] = {}
        for node_type in self._node_types_by_type.values():
            if node_type.is_abstract:
                abscls = node_type.as_type(
                    as_class_name=as_class_name,
                    base=Node,
                    mixins=mixins,
                    extra=self._extra,
                )
                abscls_name = as_class_name(node_type.type_name)
                setattr(self, abscls_name, abscls)
                for subtype in node_type.subtypes:
                    if subtype.named:
                        if subtype.type_name not in self._node_bases_by_type:
                            self._node_bases_by_type[subtype.type_name] = []
                        self._node_bases_by_type[subtype.type_name].append(abscls)

        # Dictionary of dataclasses
        self._node_classes_by_type: dict[str, type[Node]] = {}
        for node_type in self._node_types_by_type.values():
            bases = self._node_bases_by_type.get(node_type.type_name, ())
            cls = node_type.as_type(
                mixins=tuple((*bases, *mixins)),
                as_class_name=as_class_name,
                extra=self._extra,
                **dataclass_kwargs,
            )
            self._node_classes_by_type[node_type.type_name] = cls
            setattr(self, as_class_name(node_type.type_name), cls)
